@@ -1,14 +1,24 @@
 import numpy as np
 import numba
 import scipy.sparse
-from functools import wraps
+from functools import wraps, reduce
 from itertools import chain
+from operator import mul
 from collections.abc import Iterable
+from scipy.sparse import spmatrix
+from numba import literal_unroll
+import warnings
 
 from ._sparse_array import SparseArray
-from ._utils import check_compressed_axes, normalize_axis, check_zero_fill_value
+from ._utils import (
+    check_compressed_axes,
+    normalize_axis,
+    check_zero_fill_value,
+    equivalent,
+    _zero_of_dtype,
+)
 
-from ._coo.umath import elemwise
+from ._umath import elemwise
 from ._coo.common import (
     clip,
     triu,
@@ -33,23 +43,67 @@ from ._coo.common import (
 )
 
 
+@numba.njit
+def nan_check(*args):
+    """
+    Check for the NaN values in Numpy Arrays
+
+    Parameters
+    ----------
+    Union[Numpy Array, Integer, Float]
+
+    Returns
+    -------
+    Boolean Whether Numpy Array Contains NaN
+
+    """
+    for i in literal_unroll(args):
+        ia = np.asarray(i)
+        if ia.size != 0 and np.isnan(np.min(ia)):
+            return True
+    return False
+
+
+def check_class_nan(test):
+    """
+    Check NaN for Sparse Arrays
+
+    Parameters
+    ----------
+    test : Union[sparse.COO, sparse.GCXS, scipy.sparse.spmatrix, Numpy Ndarrays]
+
+    Returns
+    -------
+    Boolean Whether Sparse Array Contains NaN
+
+    """
+    from ._compressed import GCXS
+    from ._coo import COO
+
+    if isinstance(test, (GCXS, COO)):
+        return nan_check(test.fill_value, test.data)
+    elif isinstance(test, spmatrix):
+        return nan_check(test.data)
+    else:
+        return nan_check(test)
+
+
 def tensordot(a, b, axes=2, *, return_type=None):
     """
     Perform the equivalent of :obj:`numpy.tensordot`.
 
     Parameters
     ----------
-    a, b : Union[COO, np.ndarray, scipy.sparse.spmatrix]
+    a, b : Union[SparseArray, np.ndarray, scipy.sparse.spmatrix]
         The arrays to perform the :code:`tensordot` operation on.
     axes : tuple[Union[int, tuple[int], Union[int, tuple[int]], optional
         The axes to match when performing the sum.
     return_type : {None, COO, np.ndarray}, optional
         Type of returned array.
 
-
     Returns
     -------
-    Union[COO, numpy.ndarray]
+    Union[SparseArray, numpy.ndarray]
         The result of the operation.
 
     Raises
@@ -151,12 +205,12 @@ def matmul(a, b):
 
     Parameters
     ----------
-    a, b : Union[COO, np.ndarray, scipy.sparse.spmatrix]
+    a, b : Union[SparseArray, np.ndarray, scipy.sparse.spmatrix]
         The arrays to perform the :code:`matmul` operation on.
 
     Returns
     -------
-    Union[COO, numpy.ndarray]
+    Union[SparseArray, numpy.ndarray]
         The result of the operation.
 
     Raises
@@ -175,6 +229,11 @@ def matmul(a, b):
             "Cannot perform dot product on types %s, %s" % (type(a), type(b))
         )
 
+    if check_class_nan(a) or check_class_nan(b):
+        warnings.warn(
+            "Nan will not be propagated in matrix multiplication", RuntimeWarning
+        )
+
     # When b is 2-d, it is equivalent to dot
     if b.ndim <= 2:
         return dot(a, b)
@@ -229,12 +288,12 @@ def dot(a, b):
 
     Parameters
     ----------
-    a, b : Union[COO, np.ndarray, scipy.sparse.spmatrix]
+    a, b : Union[SparseArray, np.ndarray, scipy.sparse.spmatrix]
         The arrays to perform the :code:`dot` operation on.
 
     Returns
     -------
-    Union[COO, numpy.ndarray]
+    Union[SparseArray, numpy.ndarray]
         The result of the operation.
 
     Raises
@@ -274,20 +333,6 @@ def _dot(a, b, return_type=None):
     from ._compressed.convert import uncompress_dimension
     from ._sparse_array import SparseArray
 
-    def sort_indices(data, indices, indptr, shape):
-        """
-        Several of the dot algorithms produce indices that
-        are out of order. So we have to do a sort of indices
-        and data.
-        """
-        coords = np.empty((2, indices.shape[0]), dtype=np.intp)
-        coords[0, :] = uncompress_dimension(indptr)
-        coords[1, :] = indices
-        linear = linear_loc(coords, shape)
-        order = np.argsort(linear, kind="mergesort")
-        indices = indices[order]
-        data = data[order]
-
     out_shape = (a.shape[0], b.shape[1])
     if all(isinstance(arr, SparseArray) for arr in [a, b]) and any(
         isinstance(arr, GCXS) for arr in [a, b]
@@ -306,7 +351,6 @@ def _dot(a, b, return_type=None):
             data, indices, indptr = _dot_csr_csr_type(a.dtype, b.dtype)(
                 out_shape, a.data, b.data, a.indices, b.indices, a.indptr, b.indptr
             )
-            sort_indices(data, indices, indptr, out_shape)
         elif a.compressed_axes == (1,):  # csc @ csc
             # a @ b = (b.T @ a.T).T
             compressed_axes = (1,)
@@ -319,9 +363,11 @@ def _dot(a, b, return_type=None):
                 b.indptr,
                 a.indptr,
             )
-            sort_indices(data, indices, indptr, out_shape[::-1])
         out = GCXS(
-            (data, indices, indptr), shape=out_shape, compressed_axes=compressed_axes
+            (data, indices, indptr),
+            shape=out_shape,
+            compressed_axes=compressed_axes,
+            prune=True,
         )
         if return_type == np.ndarray:
             return out.todense()
@@ -338,7 +384,12 @@ def _dot(a, b, return_type=None):
             data, indices, indptr = _dot_csr_ndarray_type_sparse(a.dtype, b.dtype)(
                 out_shape, a.data, a.indices, a.indptr, b
             )
-            out = GCXS((data, indices, indptr), shape=out_shape, compressed_axes=(0,))
+            out = GCXS(
+                (data, indices, indptr),
+                shape=out_shape,
+                compressed_axes=(0,),
+                prune=True,
+            )
             if return_type == COO:
                 return out.tocoo()
             return out
@@ -349,10 +400,12 @@ def _dot(a, b, return_type=None):
         data, indices, indptr = _dot_csc_ndarray_type_sparse(a.dtype, b.dtype)(
             a.shape, b.shape, a.data, a.indices, a.indptr, b
         )
-        sort_indices(data, indices, indptr, out_shape[::-1])
         compressed_axes = (1,)
         out = GCXS(
-            (data, indices, indptr), shape=out_shape, compressed_axes=compressed_axes
+            (data, indices, indptr),
+            shape=out_shape,
+            compressed_axes=compressed_axes,
+            prune=True,
         )
         if return_type == COO:
             return out.tocoo()
@@ -370,7 +423,12 @@ def _dot(a, b, return_type=None):
             data, indices, indptr = _dot_csc_ndarray_type_sparse(bt.dtype, at.dtype)(
                 bt.shape, at.shape, bt.data, b.indices, b.indptr, at
             )
-            out = GCXS((data, indices, indptr), shape=out_shape, compressed_axes=(0,))
+            out = GCXS(
+                (data, indices, indptr),
+                shape=out_shape,
+                compressed_axes=(0,),
+                prune=True,
+            )
             if return_type == COO:
                 return out.tocoo()
             return out
@@ -383,7 +441,9 @@ def _dot(a, b, return_type=None):
         data, indices, indptr = _dot_csr_ndarray_type_sparse(bt.dtype, at.dtype)(
             out_shape[::-1], bt.data, bt.indices, bt.indptr, at
         )
-        out = GCXS((data, indices, indptr), shape=out_shape, compressed_axes=(1,))
+        out = GCXS(
+            (data, indices, indptr), shape=out_shape, compressed_axes=(1,), prune=True
+        )
         if return_type == COO:
             return out.tocoo()
         return out
@@ -431,21 +491,28 @@ def _dot(a, b, return_type=None):
         return out
 
     if isinstance(a, np.ndarray) and isinstance(b, COO):
-        b = b.T
         a = a.view(type=np.ndarray)
 
         if return_type is None or return_type == np.ndarray:
             return _dot_ndarray_coo_type(a.dtype, b.dtype)(
                 a, b.coords, b.data, out_shape
             )
+        b = b.T
         coords, data = _dot_ndarray_coo_type_sparse(a.dtype, b.dtype)(
             a, b.coords, b.data, out_shape
         )
-        out = COO(coords, data, shape=out_shape, has_duplicates=False, sorted=True)
+        out = COO(
+            coords, data, shape=out_shape, has_duplicates=False, sorted=True, prune=True
+        )
         if return_type == GCXS:
             return out.asformat("gcxs")
         return out
 
+    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
+        return np.dot(a, b)
+
+    raise TypeError("Unsupported types.")
+
 
 def _memoize_dtype(f):
     """
@@ -494,18 +561,13 @@ def _csr_csr_count_nnz(
     with compressed rows: (a @ b).nnz.
 
     Parameters
-        ----------
-        out_shape : tuple
-            The shape of the output array.
-
-        indptr : ndarray
-            The empty index pointer array for the output.
-
-        a_indices, a_indptr : np.ndarray
-            The indices and index pointer array of ``a``.
-
-        b_data, b_indices, b_indptr : np.ndarray
-            The indices and index pointer array of ``b``.
+    ----------
+    out_shape : tuple
+        The shape of the output array.
+    a_indices, a_indptr : np.ndarray
+        The indices and index pointer array of ``a``.
+    b_data, b_indices, b_indptr : np.ndarray
+        The indices and index pointer array of ``b``.
     """
     n_row, n_col = out_shape
     nnz = 0
@@ -531,18 +593,15 @@ def _csr_ndarray_count_nnz(
     numpy array: (a @ b).nnz.
 
     Parameters
-        ----------
-        out_shape : tuple
-            The shape of the output array.
-
-        indptr : ndarray
-            The empty index pointer array for the output.
-
-        a_indices, a_indptr : np.ndarray
-            The indices and index pointer array of ``a``.
-
-        b : np.ndarray
-            The second input array ``b``.
+    ----------
+    out_shape : tuple
+        The shape of the output array.
+    indptr : ndarray
+        The empty index pointer array for the output.
+    a_indices, a_indptr : np.ndarray
+        The indices and index pointer array of ``a``.
+    b : np.ndarray
+        The second input array ``b``.
     """
     nnz = 0
     for i in range(out_shape[0]):
@@ -566,18 +625,15 @@ def _csc_ndarray_count_nnz(
     numpy array: (a @ b).nnz.
 
     Parameters
-        ----------
-        a_shape, b_shape : tuple
-            The shapes of the input arrays.
-
-        indptr : ndarray
-            The empty index pointer array for the output.
-
-        a_indices, a_indptr : np.ndarray
-            The indices and index pointer array of ``a``.
-
-        b : np.ndarray
-            The second input array ``b``.
+    ----------
+    a_shape, b_shape : tuple
+        The shapes of the input arrays.
+    indptr : ndarray
+        The empty index pointer array for the output.
+    a_indices, a_indptr : np.ndarray
+        The indices and index pointer array of ``a``.
+    b : np.ndarray
+        The second input array ``b``.
     """
     nnz = 0
     mask = np.full(a_shape[0], -1)
@@ -593,9 +649,13 @@ def _csc_ndarray_count_nnz(
     return nnz
 
 
+def _dot_dtype(dt1, dt2):
+    return (np.zeros((), dtype=dt1) * np.zeros((), dtype=dt2)).dtype
+
+
 @_memoize_dtype
 def _dot_csr_csr_type(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(
         nopython=True,
@@ -613,10 +673,8 @@ def _dot_csr_csr_type(dt1, dt2):
         ----------
         out_shape : tuple
             The shape of the output array.
-
         a_data, a_indices, a_indptr : np.ndarray
             The data, indices, and index pointer arrays of ``a``.
-
         b_data, b_indices, b_indptr : np.ndarray
             The data, indices, and index pointer arrays of ``b``.
         """
@@ -632,12 +690,13 @@ def _dot_csr_csr_type(dt1, dt2):
         indices = np.empty(nnz, dtype=np.intp)
         data = np.empty(nnz, dtype=dtr)
         next_ = np.full(n_col, -1)
-        sums = np.zeros(n_col)
+        sums = np.zeros(n_col, dtype=dtr)
         nnz = 0
 
         for i in range(n_row):
             head = -2
             length = 0
+            next_[:] = -1
             for j, av in zip(
                 a_indices[a_indptr[i] : a_indptr[i + 1]],
                 a_data[a_indptr[i] : a_indptr[i + 1]],
@@ -653,7 +712,7 @@ def _dot_csr_csr_type(dt1, dt2):
                         length += 1
 
             for _ in range(length):
-                if sums[head] != 0:
+                if next_[head] != -1:
                     indices[nnz] = head
                     data[nnz] = sums[head]
                     nnz += 1
@@ -672,7 +731,7 @@ def _dot_csr_csr_type(dt1, dt2):
 
 @_memoize_dtype
 def _dot_csr_ndarray_type(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(
         nopython=True,
@@ -689,10 +748,8 @@ def _dot_csr_ndarray_type(dt1, dt2):
         ----------
         a_data, a_indices, a_indptr : np.ndarray
             The data, indices, and index pointers of ``a``.
-
         b : np.ndarray
             The second input array ``b``.
-
         out_shape : Tuple[int]
             The shape of the output array.
         """
@@ -712,7 +769,7 @@ def _dot_csr_ndarray_type(dt1, dt2):
 
 @_memoize_dtype
 def _dot_csr_ndarray_type_sparse(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(
         nopython=True,
@@ -731,10 +788,8 @@ def _dot_csr_ndarray_type_sparse(dt1, dt2):
         ----------
         a_data, a_indices, a_indptr : np.ndarray
             The data, indices, and index pointers of ``a``.
-
         b : np.ndarray
             The second input array ``b``.
-
         out_shape : Tuple[int]
             The shape of the output array.
         """
@@ -765,7 +820,7 @@ def _dot_csr_ndarray_type_sparse(dt1, dt2):
 
 @_memoize_dtype
 def _dot_csc_ndarray_type_sparse(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(
         nopython=True,
@@ -784,10 +839,8 @@ def _dot_csc_ndarray_type_sparse(dt1, dt2):
         ----------
         a_data, a_indices, a_indptr : np.ndarray
             The data, indices, and index pointers of ``a``.
-
         b : np.ndarray
             The second input array ``b``.
-
         a_shape, b_shape : Tuple[int]
             The shapes of the input arrays.
         """
@@ -831,7 +884,7 @@ def _dot_csc_ndarray_type_sparse(dt1, dt2):
 
 @_memoize_dtype
 def _dot_csc_ndarray_type(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(
         nopython=True,
@@ -850,10 +903,8 @@ def _dot_csc_ndarray_type(dt1, dt2):
         ----------
         a_data, a_indices, a_indptr : np.ndarray
             The data, indices, and index pointers of ``a``.
-
         b : np.ndarray
             The second input array ``b``.
-
         a_shape, b_shape : Tuple[int]
             The shapes of the input arrays.
         """
@@ -869,7 +920,7 @@ def _dot_csc_ndarray_type(dt1, dt2):
 
 @_memoize_dtype
 def _dot_ndarray_csc_type(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(
         nopython=True,
@@ -885,10 +936,8 @@ def _dot_ndarray_csc_type(dt1, dt2):
         ----------
         a : np.ndarray
             The input array ``a``.
-
         b_data, b_indices, b_indptr : np.ndarray
             The data, indices, and index pointers of ``b``.
-
         out_shape : Tuple[int]
             The shape of the output array.
         """
@@ -906,7 +955,7 @@ def _dot_ndarray_csc_type(dt1, dt2):
 
 @_memoize_dtype
 def _dot_coo_coo_type(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(
         nopython=True,
@@ -924,10 +973,8 @@ def _dot_coo_coo_type(dt1, dt2):
         ----------
         a_shape, b_shape : tuple
             The shapes of the input arrays.
-
         a_data, a_coords : np.ndarray
             The data and coordinates of ``a``.
-
         b_data, b_coords : np.ndarray
             The data and coordinates of ``b``.
         """
@@ -943,7 +990,7 @@ def _dot_coo_coo_type(dt1, dt2):
         coords = np.empty((2, nnz), dtype=np.intp)
         data = np.empty(nnz, dtype=dtr)
         next_ = np.full(n_col, -1)
-        sums = np.zeros(n_col)
+        sums = np.zeros(n_col, dtype=dtr)
         nnz = 0
 
         for i in range(n_row):
@@ -985,7 +1032,7 @@ def _dot_coo_coo_type(dt1, dt2):
 
 @_memoize_dtype
 def _dot_coo_ndarray_type(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(nopython=True, nogil=True)
     def _dot_coo_ndarray(coords1, data1, array2, out_shape):  # pragma: no cover
@@ -998,10 +1045,8 @@ def _dot_coo_ndarray_type(dt1, dt2):
         ----------
         data1, coords1 : np.ndarray
             The data and coordinates of ``s1``.
-
         array2 : np.ndarray
             The second input array ``x2``.
-
         out_shape : Tuple[int]
             The output shape.
         """
@@ -1025,7 +1070,7 @@ def _dot_coo_ndarray_type(dt1, dt2):
 
 @_memoize_dtype
 def _dot_coo_ndarray_type_sparse(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(
         nopython=True,
@@ -1042,10 +1087,8 @@ def _dot_coo_ndarray_type_sparse(dt1, dt2):
         ----------
         data1, coords1 : np.ndarray
             The data and coordinates of ``s1``.
-
         array2 : np.ndarray
             The second input array ``x2``.
-
         out_shape : Tuple[int]
             The output shape.
         """
@@ -1085,7 +1128,7 @@ def _dot_coo_ndarray_type_sparse(dt1, dt2):
 
 @_memoize_dtype
 def _dot_ndarray_coo_type(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(nopython=True, nogil=True)
     def _dot_ndarray_coo(array1, coords2, data2, out_shape):  # pragma: no cover
@@ -1097,10 +1140,8 @@ def _dot_ndarray_coo_type(dt1, dt2):
         ----------
         array1 : np.ndarray
             The input array ``x1``.
-
         data2, coords2 : np.ndarray
             The data and coordinates of ``s2``.
-
         out_shape : Tuple[int]
             The output shape.
         """
@@ -1108,8 +1149,8 @@ def _dot_ndarray_coo_type(dt1, dt2):
 
         for oidx1 in range(out_shape[0]):
             for didx2 in range(len(data2)):
-                oidx2 = coords2[0, didx2]
-                out[oidx1, oidx2] += array1[oidx1, coords2[1, didx2]] * data2[didx2]
+                oidx2 = coords2[1, didx2]
+                out[oidx1, oidx2] += array1[oidx1, coords2[0, didx2]] * data2[didx2]
 
         return out
 
@@ -1118,7 +1159,7 @@ def _dot_ndarray_coo_type(dt1, dt2):
 
 @_memoize_dtype
 def _dot_ndarray_coo_type_sparse(dt1, dt2):
-    dtr = np.result_type(dt1, dt2)
+    dtr = _dot_dtype(dt1, dt2)
 
     @numba.jit(
         nopython=True,
@@ -1134,10 +1175,8 @@ def _dot_ndarray_coo_type_sparse(dt1, dt2):
         ----------
         array1 : np.ndarray
             The input array ``x1``.
-
         data2, coords2 : np.ndarray
             The data and coordinates of ``s2``.
-
         out_shape : Tuple[int]
             The output shape.
         """
@@ -1173,6 +1212,165 @@ def _dot_ndarray_coo_type_sparse(dt1, dt2):
     return _dot_ndarray_coo
 
 
+def _einsum_single(lhs, rhs, operand):
+    """Perform a single term einsum, i.e. any combination of transposes, sums
+    and traces of dimensions.
+
+    Parameters
+    ----------
+    lhs : str
+        The indices of the input array.
+    rhs : str
+        The indices of the output array.
+    operand : SparseArray
+        The array to perform the einsum on.
+
+    Returns
+    -------
+    output : SparseArray
+    """
+    from ._coo import COO
+
+    if lhs == rhs:
+        if not rhs:
+            # ensure scalar output
+            return operand.sum()
+        return operand
+
+    if not isinstance(operand, SparseArray):
+        # just use numpy for dense input
+        return np.einsum(f"{lhs}->{rhs}", operand)
+
+    # else require COO for operations, but check if should convert back
+    to_output_format = getattr(operand, "from_coo", lambda x: x)
+    operand = asCOO(operand)
+
+    # check if repeated / 'trace' indices mean we are only taking a subset
+    where = {}
+    for i, ix in enumerate(lhs):
+        where.setdefault(ix, []).append(i)
+
+    selector = None
+    for ix, locs in where.items():
+        loc0, *rlocs = locs
+        if rlocs:
+            # repeated index
+            if len({operand.shape[loc] for loc in locs}) > 1:
+                raise ValueError("Repeated indices must have the same dimension.")
+
+            # only select data where all indices match
+            subselector = (operand.coords[loc0] == operand.coords[rlocs]).all(axis=0)
+            if selector is None:
+                selector = subselector
+            else:
+                selector &= subselector
+
+    # indices that are removed (i.e. not in the output / `perm`)
+    # are handled by `has_duplicates=True` below
+    perm = [lhs.index(ix) for ix in rhs]
+    new_shape = tuple(operand.shape[i] for i in perm)
+
+    # select the new COO data
+    if selector is not None:
+        new_coords = operand.coords[:, selector][perm]
+        new_data = operand.data[selector]
+    else:
+        new_coords = operand.coords[perm]
+        new_data = operand.data
+
+    if not rhs:
+        # scalar output - match numpy behaviour by not wrapping as array
+        return new_data.sum()
+
+    return to_output_format(
+        COO(new_coords, new_data, shape=new_shape, has_duplicates=True)
+    )
+
+
+def einsum(subscripts, *operands):
+    """
+    Perform the equivalent of :obj:`numpy.einsum`.
+
+    Parameters
+    ----------
+    subscripts : str
+        Specifies the subscripts for summation as comma separated list of
+        subscript labels. An implicit (classical Einstein summation)
+        calculation is performed unless the explicit indicator '->' is
+        included as well as subscript labels of the precise output form.
+    operands : sequence of SparseArray
+        These are the arrays for the operation.
+
+    Returns
+    -------
+    output : SparseArray
+        The calculation based on the Einstein summation convention.
+    """
+    check_zero_fill_value(*operands)
+
+    if "->" not in subscripts:
+        # from opt_einsum: calc the output automatically
+        lhs = subscripts
+        tmp_subscripts = lhs.replace(",", "")
+        rhs = "".join(
+            # sorted sequence of indices
+            s
+            for s in sorted(set(tmp_subscripts))
+            # that appear exactly once
+            if tmp_subscripts.count(s) == 1
+        )
+    else:
+        lhs, rhs = subscripts.split("->")
+
+    if len(operands) == 1:
+        return _einsum_single(lhs, rhs, operands[0])
+
+    # if multiple arrays: align, broadcast multiply and then use single einsum
+    # for example:
+    #     "aab,cbd->dac"
+    # we first perform single term reductions and align:
+    #     aab -> ab..
+    #     cbd -> .bcd
+    # (where dots represent broadcastable size 1 dimensions), then multiply all
+    # to form the 'minimal outer product' and do a final single term einsum:
+    #     abcd -> dac
+
+    # get ordered union of indices from all terms, indicies that only appear
+    # on a single term will be removed in the 'preparation' step below
+    terms = lhs.split(",")
+    total = {}
+    sizes = {}
+    for t, term in enumerate(terms):
+        shape = operands[t].shape
+        for ix, d in zip(term, shape):
+            if d != sizes.setdefault(ix, d):
+                raise ValueError(f"Inconsistent shape for index '{ix}'.")
+            total.setdefault(ix, set()).add(t)
+    for ix in rhs:
+        total[ix].add(-1)
+    aligned_term = "".join(ix for ix, apps in total.items() if len(apps) > 1)
+
+    # NB: if every index appears exactly twice,
+    # we could identify and dispatch to tensordot here?
+
+    parrays = []
+    for term, array in zip(terms, operands):
+        # calc the target indices for this term
+        pterm = "".join(ix for ix in aligned_term if ix in term)
+        if pterm != term:
+            # perform necessary transpose and reductions
+            array = _einsum_single(term, pterm, array)
+        # calc broadcastable shape
+        shape = tuple(
+            array.shape[pterm.index(ix)] if ix in pterm else 1 for ix in aligned_term
+        )
+        parrays.append(array.reshape(shape) if array.shape != shape else array)
+
+    aligned_array = reduce(mul, parrays)
+
+    return _einsum_single(aligned_term, rhs, aligned_array)
+
+
 def stack(arrays, axis=0, compressed_axes=None):
     """
     Stack the input arrays along the given dimension.
@@ -1200,9 +1398,9 @@ def stack(arrays, axis=0, compressed_axes=None):
     --------
     numpy.stack : NumPy equivalent function
     """
-    from ._coo import COO
+    from ._compressed import GCXS
 
-    if any(isinstance(arr, COO) for arr in arrays):
+    if not all(isinstance(arr, GCXS) for arr in arrays):
         from ._coo import stack as coo_stack
 
         return coo_stack(arrays, axis)
@@ -1239,9 +1437,9 @@ def concatenate(arrays, axis=0, compressed_axes=None):
     --------
     numpy.concatenate : NumPy equivalent function
     """
-    from ._coo import COO
+    from ._compressed import GCXS
 
-    if any(isinstance(arr, COO) for arr in arrays):
+    if not all(isinstance(arr, GCXS) for arr in arrays):
         from ._coo import concatenate as coo_concat
 
         return coo_concat(arrays, axis)
@@ -1251,7 +1449,7 @@ def concatenate(arrays, axis=0, compressed_axes=None):
         return gcxs_concat(arrays, axis, compressed_axes)
 
 
-def eye(N, M=None, k=0, dtype=float, format="coo", compressed_axes=None):
+def eye(N, M=None, k=0, dtype=float, format="coo", **kwargs):
     """Return a 2-D array in the specified format with ones on the diagonal and zeros elsewhere.
 
     Parameters
@@ -1268,8 +1466,6 @@ def eye(N, M=None, k=0, dtype=float, format="coo", compressed_axes=None):
         Data-type of the returned array.
     format : str, optional
         A format string.
-    compressed_axes : iterable, optional
-        The axes to compress if returning a GCXS array.
 
     Returns
     -------
@@ -1287,7 +1483,7 @@ def eye(N, M=None, k=0, dtype=float, format="coo", compressed_axes=None):
            [0., 0., 1.],
            [0., 0., 0.]])
     """
-    from sparse import COO
+    from ._coo import COO
 
     if M is None:
         M = N
@@ -1314,10 +1510,10 @@ def eye(N, M=None, k=0, dtype=float, format="coo", compressed_axes=None):
 
     return COO(
         coords, data=data, shape=(N, M), has_duplicates=False, sorted=True
-    ).asformat(format, compressed_axes=compressed_axes)
+    ).asformat(format, **kwargs)
 
 
-def full(shape, fill_value, dtype=None, format="coo", compressed_axes=None):
+def full(shape, fill_value, dtype=None, format="coo", order="C", **kwargs):
     """Return a SparseArray of given shape and type, filled with `fill_value`.
 
     Parameters
@@ -1333,6 +1529,10 @@ def full(shape, fill_value, dtype=None, format="coo", compressed_axes=None):
         A format string.
     compressed_axes : iterable, optional
         The axes to compress if returning a GCXS array.
+    order : {'C', None}
+        Values except these are not currently supported and raise a
+        NotImplementedError.
+
     Returns
     -------
     out : SparseArray
@@ -1353,8 +1553,8 @@ def full(shape, fill_value, dtype=None, format="coo", compressed_axes=None):
         dtype = np.array(fill_value).dtype
     if not isinstance(shape, tuple):
         shape = (shape,)
-    if compressed_axes is not None:
-        check_compressed_axes(shape, compressed_axes)
+    if order not in {"C", None}:
+        raise NotImplementedError("Currently, only 'C' and None are supported.")
     data = np.empty(0, dtype=dtype)
     coords = np.empty((len(shape), 0), dtype=np.intp)
     return COO(
@@ -1364,10 +1564,10 @@ def full(shape, fill_value, dtype=None, format="coo", compressed_axes=None):
         fill_value=fill_value,
         has_duplicates=False,
         sorted=True,
-    ).asformat(format, compressed_axes=compressed_axes)
+    ).asformat(format, **kwargs)
 
 
-def full_like(a, fill_value, dtype=None, format=None, compressed_axes=None):
+def full_like(a, fill_value, dtype=None, shape=None, format=None, **kwargs):
     """Return a full array with the same shape and type as a given array.
 
     Parameters
@@ -1395,20 +1595,22 @@ def full_like(a, fill_value, dtype=None, format=None, compressed_axes=None):
     """
     if format is None and not isinstance(a, np.ndarray):
         format = type(a).__name__.lower()
-    else:
+    elif format is None:
         format = "coo"
+
+    compressed_axes = kwargs.pop("compressed_axes", None)
     if hasattr(a, "compressed_axes") and compressed_axes is None:
         compressed_axes = a.compressed_axes
     return full(
-        a.shape,
+        a.shape if shape is None else shape,
         fill_value,
         dtype=(a.dtype if dtype is None else dtype),
         format=format,
-        compressed_axes=compressed_axes,
+        **kwargs,
     )
 
 
-def zeros(shape, dtype=float, format="coo", compressed_axes=None):
+def zeros(shape, dtype=float, format="coo", **kwargs):
     """Return a SparseArray of given shape and type, filled with zeros.
 
     Parameters
@@ -1437,14 +1639,10 @@ def zeros(shape, dtype=float, format="coo", compressed_axes=None):
     array([[0, 0],
            [0, 0]])
     """
-    if compressed_axes is not None:
-        check_compressed_axes(shape, compressed_axes)
-    return full(shape, 0, np.dtype(dtype)).asformat(
-        format, compressed_axes=compressed_axes
-    )
+    return full(shape, 0, np.dtype(dtype)).asformat(format, **kwargs)
 
 
-def zeros_like(a, dtype=None, format=None, compressed_axes=None):
+def zeros_like(a, dtype=None, shape=None, format=None, **kwargs):
     """Return a SparseArray of zeros with the same shape and type as ``a``.
 
     Parameters
@@ -1470,21 +1668,10 @@ def zeros_like(a, dtype=None, format=None, compressed_axes=None):
     array([[0, 0, 0],
            [0, 0, 0]])
     """
-    if format is None and not isinstance(a, np.ndarray):
-        format = type(a).__name__.lower()
-    elif format is None:
-        format = "coo"
-    if hasattr(a, "compressed_axes") and compressed_axes is None:
-        compressed_axes = a.compressed_axes
-    return zeros(
-        a.shape,
-        dtype=(a.dtype if dtype is None else dtype),
-        format=format,
-        compressed_axes=compressed_axes,
-    )
+    return full_like(a, 0, dtype=dtype, shape=shape, format=format, **kwargs)
 
 
-def ones(shape, dtype=float, format="coo", compressed_axes=None):
+def ones(shape, dtype=float, format="coo", **kwargs):
     """Return a SparseArray of given shape and type, filled with ones.
 
     Parameters
@@ -1513,14 +1700,10 @@ def ones(shape, dtype=float, format="coo", compressed_axes=None):
     array([[1, 1],
            [1, 1]])
     """
-    if compressed_axes is not None:
-        check_compressed_axes(shape, compressed_axes)
-    return full(shape, 1, np.dtype(dtype)).asformat(
-        format, compressed_axes=compressed_axes
-    )
+    return full(shape, 1, np.dtype(dtype)).asformat(format, **kwargs)
 
 
-def ones_like(a, dtype=None, format=None, compressed_axes=None):
+def ones_like(a, dtype=None, shape=None, format=None, **kwargs):
     """Return a SparseArray of ones with the same shape and type as ``a``.
 
     Parameters
@@ -1546,18 +1729,7 @@ def ones_like(a, dtype=None, format=None, compressed_axes=None):
     array([[1, 1, 1],
            [1, 1, 1]])
     """
-    if format is None and not isinstance(a, np.ndarray):
-        format = type(a).__name__.lower()
-    else:
-        format = "coo"
-    if hasattr(a, "compressed_axes") and compressed_axes is None:
-        compressed_axes = a.compressed_axes
-    return ones(
-        a.shape,
-        dtype=(a.dtype if dtype is None else dtype),
-        format=format,
-        compressed_axes=compressed_axes,
-    )
+    return full_like(a, 1, dtype=dtype, shape=shape, format=format, **kwargs)
 
 
 def outer(a, b, out=None):
@@ -1583,7 +1755,8 @@ def outer(a, b, out=None):
            [0, 2, 4, 6],
            [0, 3, 6, 9]])
     """
-    from sparse import SparseArray, COO
+    from ._sparse_array import SparseArray
+    from ._coo import COO
 
     if isinstance(a, SparseArray):
         a = COO(a)
@@ -1621,7 +1794,7 @@ def moveaxis(a, source, destination):
 
     Parameters
     ----------
-    a : COO
+    a : SparseArray
         The array whose axes should be reordered.
     source : int or List[int]
         Original positions of the axes to move. These must be unique.
@@ -1630,7 +1803,7 @@ def moveaxis(a, source, destination):
 
     Returns
     -------
-    COO
+    SparseArray
         Array with moved axes.
 
     Examples
@@ -1663,3 +1836,82 @@ def moveaxis(a, source, destination):
 
     result = a.transpose(order)
     return result
+
+
+def pad(array, pad_width, mode="constant", **kwargs):
+    """
+    Performs the equivalent of :obj:`numpy.pad` for :obj:`SparseArray`. Note that
+    this function returns a new array instead of a view.
+
+    Parameters
+    ----------
+    array : SparseArray
+        Sparse array which is to be padded.
+
+    pad_width : {sequence, array_like, int}
+        Number of values padded to the edges of each axis. ((before_1, after_1), … (before_N, after_N)) unique pad widths for each axis. ((before, after),) yields same before and after pad for each axis. (pad,) or int is a shortcut for before = after = pad width for all axes.
+
+    mode : str
+        Pads to a constant value which is fill value. Currently only constant mode is implemented
+
+    constant_values : int
+        The values to set the padded values for each axis. Default is 0. This must be same as fill value.
+
+    Returns
+    -------
+    SparseArray
+        The padded sparse array.
+
+    Raises
+    ------
+    NotImplementedError
+        If mode != 'constant' or there are unknown arguments.
+
+    ValueError
+        If constant_values != self.fill_value
+
+    See Also
+    --------
+    :obj:`numpy.pad` : NumPy equivalent function
+
+    """
+    if not isinstance(array, SparseArray):
+        raise NotImplementedError("Input array is not compatible.")
+
+    if mode.lower() != "constant":
+        raise NotImplementedError(f"Mode '{mode}' is not yet supported.")
+
+    if not equivalent(
+        kwargs.pop("constant_values", _zero_of_dtype(array.dtype)), array.fill_value
+    ):
+        raise ValueError("constant_values can only be equal to fill value.")
+
+    if kwargs:
+        raise NotImplementedError("Additional Unknown arguments present.")
+
+    from ._coo import COO
+
+    array = array.asformat("coo")
+
+    pad_width = np.broadcast_to(pad_width, (len(array.shape), 2))
+    new_coords = array.coords + pad_width[:, 0:1]
+    new_shape = tuple(
+        [
+            array.shape[i] + pad_width[i, 0] + pad_width[i, 1]
+            for i in range(len(array.shape))
+        ]
+    )
+    new_data = array.data
+    return COO(new_coords, new_data, new_shape, fill_value=array.fill_value)
+
+
+def format_to_string(format):
+    if isinstance(format, type):
+        if not issubclass(format, SparseArray):
+            raise ValueError(f"invalid format: {format}")
+        format = format.__name__.lower()
+
+    if isinstance(format, str):
+        return format
+
+    raise ValueError(f"invalid format: {format}")
