@@ -1,21 +1,29 @@
+import copy as _copy
 import numpy as np
+import operator
 from numpy.lib.mixins import NDArrayOperatorsMixin
 from functools import reduce
-from operator import mul
 from collections.abc import Iterable
 import scipy.sparse as ss
+from scipy.sparse import compressed
+from typing import Tuple
 
-from .._sparse_array import SparseArray
+from .._sparse_array import SparseArray, _reduce_super_ufunc
 from .._coo.common import linear_loc
 from .._common import dot, matmul
-from .._utils import normalize_axis, check_zero_fill_value, check_compressed_axes
+from .._utils import (
+    normalize_axis,
+    can_store,
+    check_zero_fill_value,
+    check_compressed_axes,
+    equivalent,
+)
 from .._coo.core import COO
 from .convert import uncompress_dimension, _transpose, _1d_reshape
 from .indexing import getitem
 
 
-def _from_coo(x, compressed_axes=None):
-
+def _from_coo(x, compressed_axes=None, idx_dtype=None):
     if x.ndim == 0:
         if compressed_axes is not None:
             raise ValueError("no axes to compress for 0d array")
@@ -43,16 +51,30 @@ def _from_coo(x, compressed_axes=None):
     compressed_shape = (row_size, col_size)
     shape = x.shape
 
+    if idx_dtype and not can_store(idx_dtype, max(max(compressed_shape), x.nnz)):
+        raise ValueError(
+            "cannot store array with the compressed shape {} and nnz {} with dtype {}.".format(
+                compressed_shape,
+                x.nnz,
+                idx_dtype,
+            )
+        )
+
+    if not idx_dtype:
+        idx_dtype = x.coords.dtype
+        if not can_store(idx_dtype, max(max(compressed_shape), x.nnz)):
+            idx_dtype = np.min_scalar_type(max(max(compressed_shape), x.nnz))
+
     # transpose axes, linearize, reshape, and compress
     linear = linear_loc(x.coords[axis_order], reordered_shape)
     order = np.argsort(linear)
     linear = linear[order]
-    coords = np.empty((2, x.nnz), dtype=np.intp)
+    coords = np.empty((2, x.nnz), dtype=idx_dtype)
     strides = 1
     for i, d in enumerate(compressed_shape[::-1]):
         coords[-(i + 1), :] = (linear // strides) % d
         strides *= d
-    indptr = np.empty(row_size + 1, dtype=np.intp)
+    indptr = np.empty(row_size + 1, dtype=idx_dtype)
     indptr[0] = 0
     np.cumsum(np.bincount(coords[0], minlength=row_size), out=indptr[1:])
     indices = coords[1]
@@ -66,11 +88,21 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
 
     This is stored in GCXS format, a generalization of the GCRS/GCCS formats
     from 'Efficient storage scheme for n-dimensional sparse array: GCRS/GCCS':
-    https://ieeexplore.ieee.org/document/7237032. GCXS generalizes the csr/csc
-    sparse matrix formats. For arrays with ndim == 2, GCXS is the same csr/csc.
+    https://ieeexplore.ieee.org/document/7237032. GCXS generalizes the CRS/CCS
+    sparse matrix formats.
+
+    For arrays with ndim == 2, GCXS is the same CSR/CSC.
     For arrays with ndim >2, any combination of axes can be compressed,
     significantly reducing storage.
 
+    GCXS consists of 3 arrays. Let the 3 arrays be RO, CO and VL. The first element
+    of array RO is the integer 0 and later elements are the number of
+    cumulative non-zero elements in each row for GCRS, column for
+    GCCS. CO stores column indexes of non-zero elements at each row for GCRS, column for GCCS.
+    VL stores the values of the non-zero array elements.
+
+    The superiority of the GCRS/GCCS over traditional (CRS/CCS) is shown by both
+    theoretical analysis and experimental results, outlined in the linked research paper.
 
     Parameters
     ----------
@@ -81,6 +113,9 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
         The shape of the array.
     compressed_axes : Iterable[int]
         The axes to compress.
+    prune : bool, optional
+        A flag indicating whether or not we should prune any fill-values present in
+        the data array.
     fill_value: scalar, optional
         The fill value for this array.
 
@@ -102,7 +137,17 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
 
     __array_priority__ = 12
 
-    def __init__(self, arg, shape=None, compressed_axes=None, fill_value=0):
+    def __init__(
+        self,
+        arg,
+        shape=None,
+        compressed_axes=None,
+        prune=False,
+        fill_value=0,
+        idx_dtype=None,
+    ):
+        if isinstance(arg, ss.spmatrix):
+            arg = self.from_scipy_sparse(arg)
 
         if isinstance(arg, np.ndarray):
             (arg, shape, compressed_axes, fill_value) = _from_coo(
@@ -110,7 +155,19 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
             )
 
         elif isinstance(arg, COO):
-            (arg, shape, compressed_axes, fill_value) = _from_coo(arg, compressed_axes)
+            (arg, shape, compressed_axes, fill_value) = _from_coo(
+                arg, compressed_axes, idx_dtype
+            )
+
+        elif isinstance(arg, GCXS):
+            if compressed_axes is not None and arg.compressed_axes != compressed_axes:
+                arg = arg.change_compressed_axes(compressed_axes)
+            (arg, shape, compressed_axes, fill_value) = (
+                (arg.data, arg.indices, arg.indptr),
+                arg.shape,
+                arg.compressed_axes,
+                arg.fill_value,
+            )
 
         if shape is None:
             raise ValueError("missing `shape` argument")
@@ -126,19 +183,36 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
             raise ValueError("data must be a scalar or 1-dimensional.")
 
         self.shape = shape
-        self.compressed_axes = (
+
+        self._compressed_axes = (
             tuple(compressed_axes) if isinstance(compressed_axes, Iterable) else None
         )
         self.fill_value = fill_value
 
-    @classmethod
-    def from_numpy(cls, x, compressed_axes=None, fill_value=0):
-        coo = COO(x, fill_value=fill_value)
-        return cls.from_coo(coo, compressed_axes)
+        if prune:
+            self._prune()
+
+    def copy(self, deep=True):
+        """Return a copy of the array.
+
+        Parameters
+        ----------
+        deep : boolean, optional
+            If True (default), the internal coords and data arrays are also
+            copied. Set to ``False`` to only make a shallow copy.
+        """
+        return _copy.deepcopy(self) if deep else _copy.copy(self)
 
     @classmethod
-    def from_coo(cls, x, compressed_axes=None):
-        (arg, shape, compressed_axes, fill_value) = _from_coo(x, compressed_axes)
+    def from_numpy(cls, x, compressed_axes=None, fill_value=0, idx_dtype=None):
+        coo = COO(x, fill_value=fill_value, idx_dtype=idx_dtype)
+        return cls.from_coo(coo, compressed_axes, idx_dtype)
+
+    @classmethod
+    def from_coo(cls, x, compressed_axes=None, idx_dtype=None):
+        (arg, shape, compressed_axes, fill_value) = _from_coo(
+            x, compressed_axes, idx_dtype
+        )
         return cls(
             arg, shape=shape, compressed_axes=compressed_axes, fill_value=fill_value
         )
@@ -156,9 +230,13 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
             )
 
     @classmethod
-    def from_iter(cls, x, shape=None, compressed_axes=None, fill_value=None):
+    def from_iter(
+        cls, x, shape=None, compressed_axes=None, fill_value=None, idx_dtype=None
+    ):
         return cls.from_coo(
-            COO.from_iter(x, shape, fill_value), compressed_axes=compressed_axes
+            COO.from_iter(x, shape, fill_value),
+            compressed_axes,
+            idx_dtype,
         )
 
     @property
@@ -197,6 +275,29 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
         """
         return self.data.shape[0]
 
+    @property
+    def format(self):
+        """
+        The storage format of this array.
+        Returns
+        -------
+        str
+            The storage format of this array.
+        See Also
+        -------
+        scipy.sparse.dok_matrix.format : The Scipy equivalent property.
+        Examples
+        -------
+        >>> import sparse
+        >>> s = sparse.random((5,5), density=0.2, format='dok')
+        >>> s.format
+        'dok'
+        >>> t = sparse.random((5,5), density=0.2, format='coo')
+        >>> t.format
+        'coo'
+        """
+        return "gcxs"
+
     @property
     def nbytes(self):
         """
@@ -251,6 +352,40 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
 
     __getitem__ = getitem
 
+    def _reduce_calc(self, method, axis, keepdims=False, **kwargs):
+        if axis[0] is None or np.array_equal(axis, np.arange(self.ndim, dtype=np.intp)):
+            x = self.flatten().tocoo()
+            out = x.reduce(method, axis=None, keepdims=keepdims, **kwargs)
+            if keepdims:
+                return (out.reshape(np.ones(self.ndim, dtype=np.intp)),)
+            return (out,)
+
+        r = np.arange(self.ndim, dtype=np.intp)
+        compressed_axes = [a for a in r if a not in set(axis)]
+        x = self.change_compressed_axes(compressed_axes)
+        idx = np.diff(x.indptr) != 0
+        indptr = x.indptr[:-1][idx]
+        indices = (np.arange(x._compressed_shape[0], dtype=self.indptr.dtype))[idx]
+        data = method.reduceat(x.data, indptr, **kwargs)
+        counts = x.indptr[1:][idx] - x.indptr[:-1][idx]
+        arr_attrs = (x, compressed_axes, indices)
+        n_cols = x._compressed_shape[1]
+        return (data, counts, axis, n_cols, arr_attrs)
+
+    def _reduce_return(self, data, arr_attrs, result_fill_value):
+        x, compressed_axes, indices = arr_attrs
+        # prune data
+        mask = ~equivalent(data, result_fill_value)
+        data = data[mask]
+        indices = indices[mask]
+        out = GCXS(
+            (data, indices, []),
+            shape=(x._compressed_shape[0],),
+            fill_value=result_fill_value,
+            compressed_axes=None,
+        )
+        return out.reshape(tuple(self.shape[d] for d in compressed_axes))
+
     def change_compressed_axes(self, new_compressed_axes):
         """
         Returns a new array with specified compressed axes. This operation is similar to converting
@@ -261,6 +396,9 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
         GCXS
             A new instance of the input array with compression along the specified dimensions.
         """
+        if new_compressed_axes == self.compressed_axes:
+            return self
+
         if self.ndim == 1:
             raise NotImplementedError("no axes to compress for 1d array")
 
@@ -297,7 +435,7 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
         """
         if self.ndim == 0:
             return COO(
-                np.array([])[None],
+                np.array([]),
                 self.data,
                 shape=self.shape,
                 fill_value=self.fill_value,
@@ -351,8 +489,7 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
         return self.tocoo().todense()
 
     def todok(self):
-
-        from ..dok import DOK
+        from .. import DOK
 
         return DOK.from_coo(self.tocoo())  # probably a temporary solution
 
@@ -387,33 +524,48 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
                 (self.data, self.indices, self.indptr), shape=self.shape
             )
 
-    def asformat(self, format, compressed_axes=None):
+    def asformat(self, format, **kwargs):
         """
         Convert this sparse array to a given format.
         Parameters
         ----------
         format : str
             A format string.
+
         Returns
         -------
         out : SparseArray
             The converted array.
+
         Raises
         ------
         NotImplementedError
             If the format isn't supported.
         """
+        from .._utils import convert_format
+
+        format = convert_format(format)
+        ret = None
 
         if format == "coo":
-            return self.tocoo()
+            ret = self.tocoo()
         elif format == "dok":
-            return self.todok()
+            ret = self.todok()
+        elif format == "csr":
+            ret = CSR(self)
+        elif format == "csc":
+            ret = CSC(self)
         elif format == "gcxs":
-            if compressed_axes is None:
-                compressed_axes = self.compressed_axes
+            compressed_axes = kwargs.pop("compressed_axes", self.compressed_axes)
             return self.change_compressed_axes(compressed_axes)
 
-        raise NotImplementedError("The given format is not supported.")
+        if len(kwargs) != 0:
+            raise TypeError(f"Invalid keyword arguments provided: {kwargs}")
+
+        if ret is None:
+            raise NotImplementedError(f"The given format is not supported: {format}")
+
+        return ret
 
     def maybe_densify(self, max_size=1000, min_density=0.25):
         """
@@ -468,6 +620,7 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
     def reshape(self, shape, order="C", compressed_axes=None):
         """
         Returns a new :obj:`GCXS` array that is a reshaped version of this array.
+
         Parameters
         ----------
         shape : tuple[int]
@@ -475,14 +628,17 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
         compressed_axes : Iterable[int], optional
             The axes to compress to store the array. Finds the most efficient storage
             by default.
+
         Returns
         -------
         GCXS
             The reshaped output array.
+
         See Also
         --------
         numpy.ndarray.reshape : The equivalent Numpy function.
-        sparse.COO.reshape: The equivalent COO function.
+        sparse.COO.reshape : The equivalent COO function.
+
         Notes
         -----
         The :code:`order` parameter is provided just for compatibility with
@@ -502,7 +658,7 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
         if self.shape == shape:
             return self
 
-        if self.size != reduce(mul, shape, 1):
+        if self.size != reduce(operator.mul, shape, 1):
             raise ValueError(
                 "cannot reshape array of size {} into shape {}".format(self.size, shape)
             )
@@ -528,50 +684,9 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
             fill_value=self.fill_value,
         )
 
-    def resize(self, *args, refcheck=True, compressed_axes=None):
-        """
-        This method changes the shape and size of an array in-place.
-
-        Parameters
-        ----------
-        args : tuple, or series of integers
-            The desired shape of the output array.
-        compressed_axes : Iterable[int], optional
-            The axes to compress to store the array. Finds the most efficient storage
-            by default.
-
-        See Also
-        --------
-        numpy.ndarray.resize : The equivalent Numpy function.
-        sparse.COO.resize : The equivalent COO function.
-        """
-        from .convert import _resize
-
-        if len(args) == 1 and isinstance(args[0], tuple):
-            shape = args[0]
-        elif all(isinstance(arg, int) for arg in args):
-            shape = tuple(args)
-        else:
-            raise ValueError("Invalid input")
-
-        if any(d < 0 for d in shape):
-            raise ValueError("negative dimensions not allowed")
-
-        if self.shape == shape:
-            return
-
-        if compressed_axes is None:
-            if len(shape) == self.ndim:
-                compressed_axes = self.compressed_axes
-            elif len(shape) == 1:
-                compressed_axes = None
-            else:
-                compressed_axes = (np.argmin(shape),)
-
-        arg = _resize(self, shape, compressed_axes)
-        self.data, self.indices, self.indptr = arg
-        self.shape = shape
-        self.compressed_axes = compressed_axes
+    @property
+    def compressed_axes(self):
+        return self._compressed_axes
 
     def transpose(self, axes=None, compressed_axes=None):
         """
@@ -639,8 +754,8 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
 
         See Also
         --------
-        scipy.sparse.csr_matrix.tocsc : Scipy equivalent function.
-        scipy.sparse.csc_matrix.tocsr : Scipy equivalent function.
+        scipy.sparse.csr_matrix.transpose : Scipy equivalent function.
+        scipy.sparse.csc_matrix.transpose : Scipy equivalent function.
         numpy.ndarray.transpose : Numpy equivalent function.
         """
         if self.ndim != 2:
@@ -698,3 +813,117 @@ class GCXS(SparseArray, NDArrayOperatorsMixin):
             return matmul(other, self)
         except NotImplementedError:
             return NotImplemented
+
+    def _prune(self):
+        """
+        Prunes data so that if any fill-values are present, they are removed
+        from both indices and data.
+
+        Examples
+        --------
+        >>> coords = np.array([[0, 1, 2, 3]])
+        >>> data = np.array([1, 0, 1, 2])
+        >>> s = COO(coords, data).asformat('gcxs')
+        >>> s._prune()
+        >>> s.nnz
+        3
+        """
+        mask = ~equivalent(self.data, self.fill_value)
+        self.data = self.data[mask]
+        if len(self.indptr):
+            coords = np.stack((uncompress_dimension(self.indptr), self.indices))
+            coords = coords[:, mask]
+            self.indices = coords[1]
+            row_size = self._compressed_shape[0]
+            indptr = np.empty(row_size + 1, dtype=self.indptr.dtype)
+            indptr[0] = 0
+            np.cumsum(np.bincount(coords[0], minlength=row_size), out=indptr[1:])
+            self.indptr = indptr
+        else:
+            self.indices = self.indices[mask]
+
+
+class _Compressed2d(GCXS):
+    def __init__(
+        self, arg, shape=None, compressed_axes=None, prune=False, fill_value=0
+    ):
+        if not hasattr(arg, "shape") and shape is None:
+            raise ValueError("missing `shape` argument")
+        if shape is not None and hasattr(arg, "shape"):
+            raise NotImplementedError("Cannot change shape in constructor")
+        nd = len(shape if shape is not None else arg.shape)
+        if nd != 2:
+            raise ValueError(f"{type(self).__name__} must be 2-d, passed {nd}-d shape.")
+
+        super().__init__(
+            arg,
+            shape=shape,
+            compressed_axes=compressed_axes,
+            prune=prune,
+            fill_value=fill_value,
+        )
+
+    def __str__(self):
+        return "<{}: shape={}, dtype={}, nnz={}, fill_value={}>".format(
+            type(self).__name__,
+            self.shape,
+            self.dtype,
+            self.nnz,
+            self.fill_value,
+        )
+
+    __repr__ = __str__
+
+    @property
+    def ndim(self) -> int:
+        return 2
+
+
+class CSR(_Compressed2d):
+    """
+    The CSR or CRS scheme stores a n-dimensional array using n+1 one-dimensional arrays.
+    The 3 arrays are same as GCRS. The remaining n-2 arrays are for storing the indices of
+    the non-zero values of the sparse matrix. CSR is simply the transpose of CSC.
+
+    Sparse supports 2-D CSR.
+    """
+
+    def __init__(self, arg, shape=None, prune=False, fill_value=0):
+        super().__init__(arg, shape=shape, compressed_axes=(0,), fill_value=fill_value)
+
+    @classmethod
+    def from_scipy_sparse(cls, x):
+        x = x.asformat("csr", copy=False)
+        return cls((x.data, x.indices, x.indptr), shape=x.shape)
+
+    def transpose(self, axes: None = None, copy: bool = False) -> "CSC":
+        if axes is not None:
+            raise ValueError()
+        if copy:
+            self = self.copy()
+        return CSC((self.data, self.indices, self.indptr), self.shape[::-1])
+
+
+class CSC(_Compressed2d):
+    """
+    The CSC or CCS scheme stores a n-dimensional array using n+1 one-dimensional arrays.
+    The 3 arrays are same as GCCS. The remaining n-2 arrays are for storing the indices of
+    the non-zero values of the sparse matrix. CSC is simply the transpose of CSR.
+
+    Sparse supports 2-D CSC.
+    """
+
+    def __init__(self, arg, shape=None, prune=False, fill_value=0):
+        super().__init__(arg, shape=shape, compressed_axes=(1,), fill_value=fill_value)
+
+    @classmethod
+    def from_scipy_sparse(cls, x):
+        x = x.asformat("csc", copy=False)
+        return cls((x.data, x.indices, x.indptr), shape=x.shape)
+
+    def transpose(self, axes: None = None, copy: bool = False) -> CSR:
+        if axes is not None:
+            raise ValueError()
+        if copy:
+            self = self.copy()
+        return CSR((self.data, self.indices, self.indptr), self.shape[::-1])
